/////////////////////////////////////////////////////////////////////////////
// gpioshutdown - written by Bob Frazier for S.F.T. Inc. http://mrp3.com/
//                contributed to the FreeBSD source as-is, to address
//                bug report 211979
//
// This module's purpose is to reset all non-reserved GPIO pins to a default
// (input) state on shutdown, right before halt.  This way an external device
// that relies on GPIO pins being reset this way to identify a 'power off'
// condition can THEN detect a shutdown and power off the CPU, once it's safe.
//
// For more info, see bug 211979.
//
// This is primarily intended for the Raspberry Pi, but can be
// included by [or dynamically loaded into] any other kernel, as needed.

/* Portions of this were derived from existing FreeBSD kernel source, with
 * the following copyright and BSD license:
 *
 * Copyright (c) 2009 Oleksandr Tymoshenko <gonzo@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>

#include "opt_platform.h"

// NOTE: these headers were copied from other driver sources, and some may not be needed
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/bus.h>
#include <sys/gpio.h>
#include <sys/kernel.h>
#include <sys/lock.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/conf.h>
#include <sys/eventhandler.h>
#include <sys/sysctl.h>

#include <dev/gpio/gpiobusvar.h>
#include <dev/led/led.h> /* LED device */

#include "gpio_if.h"
#include "gpiobus_if.h"

//#define USE_MOD_SHUTDOWN /* uncomment to use 'MOD_SHUTDOWN' in lieu of registered event handler */

//#define DEBUG_VERBOSE /* uncomment this define this to display verbose messages */

#ifdef DEBUG_VERBOSE
#define INTERNAL_DEBUG_PRINT(...) printf(__VA_ARGS__)
#else // !DEBUG_VERBOSE
#define INTERNAL_DEBUG_PRINT(...)
#endif // DEBUG_VERBOSE

static int led_gpio = -1, led_gpio_set = 1; // LED GPIO pin and 'set' value

static void internal_do_shutdown(device_t dev, int *pnPins, int led_list[], int nLED);

// I added a handler proc so I can monitor certain things THAT way...
// however, I'm calling this from 'MOD_SHUTDOWN' for now
static int gpioshutdown_handler(module_t mod, int /*modeventtype_t*/ what, void *arg);

// NOTE:  pins are defined in /usr/src/sys/boot/fdt/dts/arm  rpi.dts, rpi2.dts
//        worth noting, the serial port is GPIO 14 (TX) and 15 (RX)  [these usually have flags == 0]
//        On 1B the 'ok' LED is 16; on Rev 2 the 'act' LED is 47
//        On Rev 2 the 'pwr' LED is 35.  The chart at https://elinux.org/RPi_BCM2835_GPIOs is wrong

static void internal_do_shutdown(device_t dev, int *pnPins, int led_list[], int nLED)
{
int max_pin, res, caps, pin, val, pin_base, ctr;

  INTERNAL_DEBUG_PRINT("gpioshutdown - internal_do_shutdown\n");

  // this is where I do all the work.  for now, just assume all non-reserved
  // GPIO pins are to be converted to inputs without pull up/down resistors,
  // and that reserved pins will fail when I check the caps against the pin


  res = GPIO_PIN_MAX(dev, &max_pin);

  pin_base = *pnPins;
  *pnPins += max_pin + 1; // note that 'max_pin' is INCLUSIVE

  for(pin=0; pin <= max_pin; pin++)
  {
    res = GPIO_PIN_GETFLAGS(dev, pin, &val);

    if(!(val & (GPIO_PIN_INPUT | GPIO_PIN_OUTPUT))) // pin is neither input nor output (don't mess with it)
    {
      INTERNAL_DEBUG_PRINT("\nskipping pin %d, flags=%08xH", pin + pin_base, val);
      continue;
    }

    // TODO:  find out whether or not this is normally a reserved pin, such as having
    //        an 'ALT' designation in the broadcom implementation...
    //        (this is to correct the apparent bug caused by all of the I/O pin definitions
    //        being assigned to 'input' within the array, but not on the actual pin)

    res = GPIO_PIN_GETCAPS(dev, pin, &caps);
    if (res == 0)
    {
      if(val & GPIO_PIN_OUTPUT)
      {
        // See if it's in my LED list.  if it is, handle differently
        for(ctr=0; ctr < nLED; ctr++)
        {
          if(led_list[ctr] == pin + pin_base)
          {
            printf("\ngpioshutdown - OUTPUT pin %d is LED", pin + pin_base);

            break;
          }
        }

        if(ctr < nLED) // LED pin found in list
        {
//          int val2 = 0;

          if(pin + pin_base == led_gpio)
          {
            GPIO_PIN_SET(dev, pin, led_gpio_set ? 0 : 1);  // assign to 'not led_gpio_set'
          }
          else
          {
            GPIO_PIN_SET(dev, pin, 0); // assume zero
          }

//          GPIO_PIN_GET(dev, pin, &val2);
//
//          if(val2 != 0)
//          {
//            printf("\ngpioshutdown - WARNING: OUTPUT pin %d is _STILL_ set to 1!", pin + pin_base);
//          }
//          else
//          {
//            printf("\ngpioshutdown - INFO: OUTPUT pin %d was set to 0", pin + pin_base);
//          }

          continue; // leaving it as an output pin with value '0'
        }
      }

      if(caps & GPIO_PIN_INPUT) // it CAN be an input - see gpio_check_flags() in gpiobus.c
      {
        res = GPIO_PIN_SETFLAGS(dev, pin, GPIO_PIN_INPUT); // change it to an input

        if(val == GPIO_PIN_INPUT)
        {
          INTERNAL_DEBUG_PRINT("\nset flags on input pin %d, result=%d", pin + pin_base, res);
        }
        else
        {
          INTERNAL_DEBUG_PRINT("set flags on output pin %d, result=%d\n", pin + pin_base, res);
        }
      }
      else
      {
        /*INTERNAL_DEBUG_PRINT*/printf("\nskipping pin %d, caps = (%xH)", pin + pin_base, caps);
      }
    }
    else
    {
      INTERNAL_DEBUG_PRINT("\nskipping pin %d, no caps", pin + pin_base);
    }
  }

  INTERNAL_DEBUG_PRINT("\ngpioshutdown - internal_do_shutdown - done, max_pin=%d\n", max_pin);
}


#ifndef USE_MOD_SHUTDOWN
static eventhandler_tag evt_tag; // uncomment if I register an event handler instead
#endif // !USE_MOD_SHUTDOWN

static eventhandler_tag evt_tag_ps;

static int internal_shutdown_pre_sync(void *arg1 __unused)
{
int nMaxUnit, nUnit, nPinBase, max_pin, res;
devclass_t dc;
device_t dev;


  printf("internal_shutdown_pre_sync - setting LEDs...");

//  led_set("pwr","1\n");
//  led_set("act","1\n");
//  led_set("ok","0\n");

  if(led_gpio < 0)
  {
    return 0; // nothing to do here
  }

  dc = devclass_find("gpio"); // we want all of the GPIO device class drivers

  if(dc)
  {
    nMaxUnit = devclass_get_maxunit(dc);
    nPinBase = 0;

    // find 'led_gpio' GPIO pin and set to 'led_gpio_set'
    // This will turn the LED on before sync so I can watch it
    // turn off to know that it's safe to power down the unit.

    for(nUnit=0; nUnit < nMaxUnit; nUnit++)
    {
      dev = devclass_get_device(dc, nUnit);

      if(dev)
      {
        max_pin = 0;

        res = GPIO_PIN_MAX(dev, &max_pin);

        if(led_gpio >= nPinBase && led_gpio <= max_pin + nPinBase)
        {
          int pin, val;

          pin = led_gpio - nPinBase;

          res = GPIO_PIN_GETFLAGS(dev, pin, &val);
          if(!(val & GPIO_PIN_OUTPUT))
          {
            int caps;
            res = GPIO_PIN_GETCAPS(dev, pin, &caps);

            if(caps & GPIO_PIN_OUTPUT) // it CAN be an input - see gpio_check_flags() in gpiobus.c
            {
              GPIO_PIN_SETFLAGS(dev, pin, GPIO_PIN_OUTPUT); // in case it's not

              goto set_output_pin;
            }
          }
          else // already an output pin
          {
set_output_pin:
            GPIO_PIN_SET(dev, pin, led_gpio_set ? 1 : 0);  // assign to 'led_gpio_set'
          }
        }

        nPinBase += max_pin + 1; // 'max_pin' is INCLUSIVE
      }
    }
  }

  printf(" complete!\n");

  return 0;
}

#define MAX_LED_LIST 64

static int internal_shutdown_final(void *arg1)
{
devclass_t dc;
device_t dev;
int nMaxUnit, nUnit;
int led_list[MAX_LED_LIST];
int nLED = 0;
int nPinsSoFar = 0;

//  INTERNAL_DEBUG_PRINT("gpioshutdown - internal_shutdown_final\n");
  printf("gpioshutdown - I/O pin reset begin ...");

  // scan for LEDs

//  dc = devclass_find("led");
//
//  if(dc)
//  {
//    nMaxUnit = devclass_get_maxunit(dc);
//    for(nUnit=0; nUnit < nMaxUnit; nUnit++)
//    {
//      dev = devclass_get_device(dc, nUnit);
//
//
//    }
//  }

  if(led_gpio >= 0)
  {
    led_list[nLED++] = led_gpio;
  }

  // scan for all gpio drivers, call internal_do_shutdown

  dc = devclass_find("gpio"); // we want all of the GPIO device class drivers

  if(dc)
  {
    nMaxUnit = devclass_get_maxunit(dc);

    for(nUnit=0; nUnit < nMaxUnit; nUnit++)
    {
      dev = devclass_get_device(dc, nUnit);

      if(dev)
      {
        INTERNAL_DEBUG_PRINT("gpioshutdown - internal_shutdown_final - temporary, doing unit %d\n", nUnit);
        internal_do_shutdown(dev, &nPinsSoFar, led_list, nLED); // turn all of the IO pins into inputs
      }
    }
  }

  // for now, always print this message so I know when it completes
  printf(" complete!\n");

  return 0;
}

static int gpioshutdown_handler(module_t mod, int what, void *arg)
{
int err=0;

  switch(what)
  {
    case MOD_LOAD:
      INTERNAL_DEBUG_PRINT("gpioshutdown_handler - module load\n");

#ifndef USE_MOD_SHUTDOWN
      evt_tag = EVENTHANDLER_REGISTER(shutdown_final, internal_shutdown_final, 0, EVENTHANDLER_PRI_LAST);
#endif // !USE_MOD_SHUTDOWN
      evt_tag_ps = EVENTHANDLER_REGISTER(shutdown_pre_sync, internal_shutdown_pre_sync, 0, EVENTHANDLER_PRI_FIRST);

      break;

    case MOD_UNLOAD:
#ifndef USE_MOD_SHUTDOWN
      EVENTHANDLER_DEREGISTER(shutdown_final, evt_tag);
#endif // !USE_MOD_SHUTDOWN
      EVENTHANDLER_DEREGISTER(shutdown_pre_sync, evt_tag_ps);

      INTERNAL_DEBUG_PRINT("gpioshutdown_handler - module unload\n");
      break;

    case MOD_SHUTDOWN:
#ifdef USE_MOD_SHUTDOWN
      internal_shutdown_final(0); // do the GPIO shutdown from the 'MOD_SHUTDOWN' event
#endif // !USE_MOD_SHUTDOWN

      INTERNAL_DEBUG_PRINT("gpioshutdown_handler - module shutdown\n");
      break;

    case MOD_QUIESCE:
      INTERNAL_DEBUG_PRINT("gpioshutdown_handler - module quiesce\n");
      break;

    default:
      INTERNAL_DEBUG_PRINT("gpioshutdown_handler - module command %d (err)\n", what);
      err = EINVAL;
      break;
  }

  return err;
}

/////////////////////////////////////////////
// SYSCTL VARIABLES - kern.gpioshutdown.xxx

// use the pre-defined _kern node and create _kern_gpioshutdown node
SYSCTL_NODE(_kern, OID_AUTO, gpioshutdown, CTLFLAG_RW, 0, "gpioshutdown kernel module");

// kern.gpioshutdown.led_gpio - integer, -1 to disable, else gpio # for LED port
// GPIO can be derived using ofwdump -P gpios /leds/pwr or ofwdump -P gpios /leds/ok  [as appropriate]
// the 2nd line of the output contains 12 hex digits, the middle 4 of which identify the GPIO pin
SYSCTL_INT(_kern_gpioshutdown, OID_AUTO, led_gpio, CTLFLAG_RW,
           &led_gpio, -1,
          "GPIO pin number to assume for LED output (set to 0 on shutdown)");
SYSCTL_INT(_kern_gpioshutdown, OID_AUTO, led_gpio_set, CTLFLAG_RW,
           &led_gpio_set, 1,
          "GPIO pin value to assign for LED \"on\"");

// a semi-simple way of extracting the GPIO pin on an RPi using ofwdump and awk:
// ofwdump -P gpios /leds/pwr | grep -v "gpios:" | awk '
//    function hexdigit(x) { xx = index("0123456789ABCDEF",toupper(x)) - 1; return xx; }
//    function hexval(x) { return hexdigit(substr(x,1,1)) * 16 + hexdigit(substr(x,2,1)); }
//    { print hexval($5) * 65536 * 256 + hexval($6) * 65536 + hexval($7) * 256 + hexval($8); }'



///////////////////////
// MODULE DECLARATION

// this is a simple module, can be dynamically loaded
DEV_MODULE(gpioshutdown, gpioshutdown_handler, 0);
MODULE_VERSION(gpioshutdown, 1);
//MODULE_DEPEND(gpioshutdown,gpiobus,1,1,1);


